# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vermouth.gmx import write_molecule_itp
import networkx as nx


def network_writer(ff, molname, bonds, ext, network_type):
    """
    write an elastic network only topology for a particular molecule

    Parameters
    ----------
    ff: vermouth forcefield
        vermouth force field containing the input system
    molname: str
        name of the molecule to separate out
    bonds: list
        list of elastic/Go network bonds to write out
    network_type: str
        should be either "Go" or "elastic" to denote the name of the network

    Returns
    -------
    None
    """
    graph = nx.Graph()
    graph.add_nodes_from(ff.blocks[molname].nodes)
    graph.add_edges_from(bonds)

    removed = []

    '''
    for each node in the graph, check its degree
    if its degree is > 12, remove the 'excess' number of edges from
    that node
    '''
    for node in graph.nodes:
        degree = graph.degree[node]
        to_remove = degree - 12
        if to_remove > 0:
            removal_list = list(graph.edges(node))[:to_remove]
            removed.append([sorted(i) for i in removal_list])
            graph.remove_edges_from(removal_list)

    all_removed = sorted([i for j in removed for i in j])

    assert max([graph.degree[node] for node in graph.nodes]) < 13

    for interaction_type in list(ff.blocks[molname].interactions):
        del ff.blocks[molname].interactions[interaction_type]

    for bond in graph.edges:
        ff.blocks[molname].add_interaction('bonds', [bond[0], bond[1]], ['1', '1', '1000'])

    if len(all_removed) > 0:
        with open(molname + f'_surplus_{network_type}.txt', 'w') as extra_en:
            if network_type == 'en':
                network_type_write = "elastic"
            extra_en.writelines(f"{network_type_write} network bonds removed from {molname}_{network_type}.itp\n"
                                "This is for noting in visualisation, not for simulation\n\n"
                                f"These bonds will be missing if you load {molname}_{network_type}.itp in vmd\n"
                                "having been present in your simulation. If you're inspecting your\n"
                                f"{network_type_write} network because you suspect some simulation error\n"
                                f"as a result of how it was constructed, bear this in mind.\n\n"
                                "This has been done because by default VMD does not draw more than 12 bonds\n"
                                "per atom in an interactive manner (i.e. the bonds are not rendered as static geometry).\n"
                                "To view all bond (in a static manner) the bonds can be rendered using the -cyl flag of MartiniGlass.")

            extra_en.write('     i      j func b0 kb\n')

            for i in all_removed:
                extra_en.writelines(f'{i[0]:6d}\t{i[1]:6d}\n')

    # write the file out
    mol_out = ff.blocks[molname].to_molecule()
    mol_out.meta['moltype'] = molname + f'_{network_type}'

    if ext:

        ext_bonds_list = [i.atoms for i in mol_out.interactions['bonds']]
        stout = ''.join([f'{i[0]}\t{i[1]}\n' for i in ext_bonds_list])
        with open(f'{molname}_{network_type}_bonds.txt', 'w') as bonds_list_out:
            bonds_list_out.write(stout)

    header = [f'{network_type} network topology for {molname}', 'NOT FOR SIMULATIONS']

    with open(molname + f'_{network_type}.itp', 'w') as fout:
        write_molecule_itp(mol_out, outfile=fout, header=header)

    return fout.name



# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
import numpy as np


def parse_fixed_width_line(line):
    # fixed widths in a gro file
    field_widths = [5, 5, 5, 5, 8, 8, 8, 8, 8, 8]  # Field widths derived from the C format specifiers
    field_types = ['int', 'str', 'str', 'int', 'float', 'float', 'float', 'float', 'float', 'float']  # Data types
    fields = []
    start = 0
    for width, f_type in zip(field_widths, field_types):
        field = line[start:start+width].strip()  # Slice the line and strip spaces
        try:
            if f_type == 'int':
                fields.append(int(field))  # Convert to integer
            elif f_type == 'float':
                fields.append(float(field))  # Convert to float
            else:
                fields.append(field)  # Leave as string
        except ValueError:
            pass
        start += width
    return fields


def get_positions(coord_file):
    with open(coord_file, 'r') as f:
        lines = [parse_fixed_width_line(i)[4:7] for i in f.readlines()[2:-1]]  # gets the coordinates
    return np.array(lines)


def file_write(coord_file, topology_contents, bonds_dictionary, mol_lens):
    """
    coord file: input coordinate file of system
    topology_contents
    bonds_dictionary: dict
        dictionary of all index based external network bonds for all molecules in system
    mol_lens: dict
        dictionary of the number of nodes of each molecule
    """
    coords = get_positions(coord_file)

    # need to loop over this list because the order the topology file was processed in matters
    mols_list = [i['name'] for i in topology_contents['molecules']]

    cylinder_string = "\n"
    start = 0
    for mol in mols_list:
        entry = [i for i in topology_contents['molecules'] if i['name'] == mol]
        n_mols = entry[0]['n_mols']

        # this is the starting indices of each molecule in the system
        start_indices = np.arange(start, start+(int(n_mols) * int(mol_lens[mol])), int(mol_lens[mol]))
        start += int(n_mols) * int(mol_lens[mol])

        cylinder_string += f"\n# starting bonds for {mol} here"
        for bond in bonds_dictionary[mol]:

            starts = np.array(start_indices+bond[0])
            ends = np.array(start_indices+bond[1])
            coord_pairs = np.stack([[i, j] for i, j in zip(coords[starts], coords[ends])])

            for position in coord_pairs:
                # need the conversion here to make it back into gromacs for some reason
                pos0_str = ' '.join([f'{j*10:.3f}' for j in position[0]])
                pos1_str = ' '.join([f'{j*10:.3f}' for j in position[1]])

                cylinder_string += f'\ngraphics 0 cylinder "{pos0_str}" "{pos1_str}" radius $radius resolution 50'

    msg = textwrap.dedent(
                            """
                            # Colour of the cylinder. 16 is black.
                            set colorID 16
                        
                            # Cylinder radius 
                            set radius 0.3
                            
                            graphics 0 color $colorID"
                        
                            # Draw a cylinder between the points listed
                            {cylinders}
                            """
                            )

    with open('network_cylinders.tcl', 'w') as f:
        f.write(textwrap.dedent(msg.format(cylinders=cylinder_string)))



# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

def index_writing(system):
    """
    Write a .ndx file for a system without water
    Parameters
    ----------
    system: str
        the .gro file to read and write the non-water index for

    Returns
    -------

    """

    if system.suffix != ".gro":
        raise TypeError('Must provide a file in .gro format')

    print("Writing a waterless index file. Here're some helpful commands for reference:\n"
          "\tgmx trjconv -f traj_comp.xtc -s topol.tpr -pbc mol -n index.ndx -e 0 -o vis.gro\n"
          "\tgmx trjconv -f traj_comp.xtc -s topol.tpr -pbc mol -n index.ndx -o vis.xtc"
          )

    # read the file, take the number of the line if it doesn't have water in
    with open(f"{system}", 'r') as file:
        lines = file.readlines()
    # this skips the header and footer of the file
    file = [j for (j, k) in enumerate(lines[2:-1], start=1) if k[10:15].strip() != "W"]
    # split the lines every 12th index as gromacs requires
    lines_out = [file[x:x + 12] for x in range(0, len(file), 12)]

    # write the index file
    with open('index.ndx', 'w') as fout:
        fout.write('[ not water ]\n')
        for lineout in lines_out:
            fout.write(' '.join(map(str, lineout)) + ' \n')



# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from os import remove

def output_file_append(frame, trajectory, proteins=False,
                       go=False, elastic=False, cylinders=False):

    if isinstance(frame, Path):
        if go:
            extra_top = "go.top"

        if elastic:
            extra_top = "en.top"

        if proteins:
            inputs = ['vis.vmd', 'proteins.vmd']
        else:
            inputs = ['vis.vmd']

        for fin in inputs:

            with open(fin) as f:
                lines = f.readlines()

            if isinstance(trajectory, Path):
                trajectory_line = f"mol addfile {trajectory} type xtc first 0 last -1 step 1 waitfor all molid 1\n"
            else:
                trajectory_line = " \n"

            if go or elastic:
                extra = (f"mol new {frame} type gro first 0 last -1 step 1\n"
                         f"{trajectory_line}"
                         f"cg_bonds -top {extra_top}\n"
                         "mol modstyle 0 1 Bonds 0.300000 52.000000\n"
                         "mol modcolor 0 1 ColorID 16\n"
                         "mol modmaterial 0 1 AOChalky\n")
            elif cylinders:
                extra = "source network_cylinders.tcl"

            lines_out = lines + [extra]

            remove(fin)

            with open(fin, "w") as fout:
                fout.writelines(lines_out)



#!/usr/bin/python3
# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import numpy as np

# need this snippet to calculate the principal axes of a set of points

if __name__ == '__main__':
    #sort out the given argument
    arrin = list(sys.argv[1].split(" "))    
    #take the input string, make into array, get eigenvectors out
    arr =  np.array(arrin[3:],dtype = float).reshape(int(arrin[1]),int(arrin[2]))
    val = np.concatenate((np.array([2,3,3]),list(np.linalg.eig(arr)[1].flatten())))
    #write to stdout
    sys.stdout.write(' '.join(str(i) for i in val))




# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

def input_topol_reader(file):
    inclusions = []
    molecules = []
    system = []
    go = []
    # read the topology file, find the header lines which are not core martini itps
    with open(file) as f:
        mols = False
        sys = False
        for line in f.readlines():
            if '#include' in line:
                inclusions.append(line.split('"')[1])
            if mols:
                if (line.strip()[0] != ';') and (len(line.split()) > 0):
                    molecules.append({'name': line.split()[0],
                                      'n_mols': line.split()[1]})
            if 'molecules' in line:
                mols = True
                sys = False
                system.append(line)
            if 'system' in line:
                sys = True
            if sys:
                system.append(line)
            if 'moleculetype' in line:
                msg = ("Your topology file contains [ moleculetype ] directives, which this file parser "
                       "cannot process. Please split your system into multiple itp files.")
                raise IOError(msg)

    topol_lines = {'core_itps': inclusions,
                   'system': system,
                   'molecules': molecules}

    return topol_lines


def topol_writing(topol_lines, written_mols, ext='vis', w_include=None):
    """

    Write new .top file based on the input one

    Parameters
    ----------
    topol_lines: dict
        lines from the input topology file split up into different keys, as per input_topol_reader
    ext: str
        the extension to use in looking for new file names to add to the output .top file
    w_include
        if W_include is not None (ie. args.system has been given something)
        then the line for water is not written out in the .top file

    Returns
    -------
    None
    """
    # this will write everything and is gross but it should work
    # because everything will be included
    new_topol_head = []
    # get the new vis files to write for the topology header
    for i in written_mols:
        new_topol_head.append(f'#include "{os.path.abspath(i)}"\n')

    # correct the [ molecules ] directive of the top file to correct for the new molecule names.
    topol_rest_vis = []
    original_mols = [i.split(f'_{ext}')[0] for i in written_mols]
    if w_include is not None:
        try:
            original_mols.remove('W')
        except ValueError:
            # print('No water to remove!')
            pass
    for i in topol_lines['molecules']:
        if any(mol in i['name'] for mol in original_mols):
            topol_rest_vis.append(i['name'] + f'_{ext}\t' + i['n_mols'] + '\n')
    # combine the sections of the vis.top and write it out.
    vis_topol = new_topol_head + topol_lines['system'] + topol_rest_vis

    with open(f'{ext}.top', 'w') as f:
        f.writelines(vis_topol)




# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from os.path import isfile
from vermouth.gmx import write_molecule_itp
from .network_writer import network_writer
from vermouth.graph_utils import make_residue_graph

def molecule_editor(ff, topol_lines,
                    virtual_sites=True, ext=False,
                    elastic=False,
                    go=False, go_path='', go_file=''):
    # iterate over the molecules to make visualisation topologies
    keep = ['bonds', 'constraints', 'pairs', 'virtual_sitesn',
            'virtual_sites2', 'virtual_sites3']
    print("Writing visualisable topology files")

    written_mols = []
    mol_bonds = {}
    mol_lens = {}

    for molname, block in ff.blocks.items():
        # write vis topols for molecules we're actually interested in
        try:
            assert molname in [i['name'] for i in topol_lines['molecules']]
        except AssertionError:
            continue

        bonds_list = []
        # delete the interactions which are not bonds
        for interaction_type in list(block.interactions):
            if interaction_type not in keep:
                del block.interactions[interaction_type]

        # remove meta (i.e. the #IFDEF FLEXIBLE) from the bonds
        for bond in block.interactions['bonds']:
            bond.meta.clear()
        if elastic:

            res_graph = make_residue_graph(block)
            res_dict = {}
            for node in res_graph.nodes:
                for bead in res_graph.nodes[node]['graph'].nodes:
                    res_dict[bead] = node

            for bond in list(block.interactions['bonds']):
                # this should introduce actual parameters into the block if the bonds have been given by
                # a #define statement elsewhere
                if len(bond.parameters) < 3:
                    atoms = bond.atoms
                    # paras = bond.parameters[0]
                    block.remove_interaction('bonds', bond.atoms)
                    # strictly the interaction here should be system_defines[paras], but it doesn't matter
                    block.add_interaction('bonds', atoms, ['1', '1', '1000'],
                                          meta={"comment": "external bond definition"})
                    continue

                else:
                    at1 = bond.atoms[0]
                    at2 = bond.atoms[1]
                    if abs(res_dict[at1] - res_dict[at2]) > 1:
                        bonds_list.append([bond.atoms[0], bond.atoms[1]])
                        block.remove_interaction('bonds', bond.atoms)

            ff_en_copy = copy.deepcopy(ff)
            en_written = network_writer(ff_en_copy, molname, bonds_list, ext, 'en')
            written_mols.append(en_written)

        # this should then keep any constraints which don't have IFDEF statements
        # e.g. alpha helices are described by constraints without these.
        # however, the remove_interactions function doesn't work atm.
        for bond in list(block.interactions['constraints']):
            if bond.meta.get('ifndef'):
                block.remove_interaction('constraints', bond.atoms)
            else:
                block.add_interaction('bonds', bond.atoms, bond.parameters + ['10000'])
                block.remove_interaction('constraints', bond.atoms)

        # rewrite pairs as bonds for visualisation
        for bond in block.interactions['pairs']:
            block.add_interaction('bonds', bond.atoms, bond.parameters[:2] + ['10000'])
        del block.interactions['pairs']

        # make bonds between virtual sites and each of the constructing atoms
        go_dict = {}
        for vs_type in ['virtual_sitesn', 'virtual_sites2', 'virtual_sites3']:
            if virtual_sites:
                for vs in block.interactions[vs_type]:
                    site = vs.atoms[0]
                    constructors = vs.atoms[1:]
                    # this avoids pointless bonds between a virtual site directly on top of
                    # its singular constructing atom
                    block.nodes[site]['mass'] = 1
                    if go:
                        # make a dictionary of atype: node index
                        # this is for later so the 'bond' can be drawn properly.
                        # assert that this site has the name CA to check it's a go site and not another VS.
                        aname = block.nodes[site]['atomname']
                        atype = block.nodes[site]['atype']
                        if (aname.split('_')[0] == 'molecule') or (aname == 'CA'):
                            go_dict[atype] = site
                    else:
                        for constructor in constructors:
                            # completely arbitrary parameters, the bond just needs to exist
                            block.add_interaction('bonds', [site, constructor],
                                                  ['1', '1', '1000'], meta={"comment": "bonded virtual site"})
            if block.interactions[vs_type]:
                del block.interactions[vs_type]

        if go:
            # len(go_dict)>0:
            if go_path:
                go_nb_file = go_file
            else:
                go_nb_file = "go_nbparams.itp"

            if not isfile(go_nb_file):
                raise FileNotFoundError("Gō nonbonded itp does not exist. Specify using -gf")

            with open(go_nb_file, "r") as f:
                nb_lines = f.readlines()
            nb_lines = [line.split(';')[0].split(' ') for line in nb_lines if '[' not in line]
            # TODO this causes problems when we're not actually in the correct block that's
            # associated with the go model! ignore the exception for now.
            try:
                for i in nb_lines:
                    try:
                        assert i[0] in go_dict
                        assert i[1] in go_dict
                        bonds_list.append([go_dict[i[0]], go_dict[i[1]]])  # , '1', i[3], '1000'])
                    except AssertionError:
                        pass
            except KeyError:
                pass
            ff_go_copy = copy.deepcopy(ff)
            go_written = network_writer(ff_go_copy, molname, bonds_list, ext, "go")
            written_mols.append(go_written)

        # write out the molecule with an amended name
        mol_out = block.to_molecule()
        mol_out.meta['moltype'] = molname + '_vis'

        if ext:

            ext_bonds_list = [i.atoms for i in mol_out.interactions['bonds']]
            stout = ''.join([f'{i[0]}\t{i[1]}\n' for i in ext_bonds_list])
            with open(f'{molname}_bonds.txt', 'w') as bonds_list_out:
                bonds_list_out.write(stout)

        header = [f'Visualisation topology for {molname}', 'NOT FOR SIMULATIONS']

        with open(molname + '_vis.itp', 'w') as fout:
            write_molecule_itp(mol_out, outfile=fout, header=header)
            written_mols.append(fout.name)

        mol_bonds[molname] = bonds_list
        mol_lens[molname] = len(block.nodes)

    return written_mols, mol_bonds, mol_lens



# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Find the data directory once.
try:
    from importlib.resources import files, as_file
    import atexit
    from contextlib import ExitStack
except ImportError:
    from pathlib import Path
    DATA_PATH = Path(__file__).parent / 'data'
    del Path
else:
    ref_data = files('martiniglass') / 'data'
    file_manager = ExitStack()
    atexit.register(file_manager.close)
    DATA_PATH = file_manager.enter_context(as_file(ref_data))

    del files, as_file, atexit, ExitStack

from .src.system_reading import system_reading
from .src.index_writer import index_writing
from .src.molecule_editing import molecule_editor
from .src.topology import topol_writing
from .src.cylinders_tcl import file_write as cylinders
from .src.output_file_append import output_file_append

from martiniglass.martiniglass import main


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
from argparse import ArgumentDefaultsHelpFormatter
from martiniglass import system_reading, index_writing, molecule_editor, topol_writing, cylinders, output_file_append
from martiniglass import DATA_PATH
import os
from pathlib import Path
import shutil

def main():

    parser = argparse.ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-p", dest="topology", type=Path, help="input .top file used", default="topol.top")
    parser.add_argument("-f", dest="system", type=Path,
                        help=("Gromacs .gro file for which to write a non-water index file. "
                              "Equivalent to an index group of !W in gmx make_ndx. "
                              "Giving this option will automatically exclude W from your output vis.top")
                        )
    parser.add_argument("-traj", dest="trajectory", type=Path,
                        help=("Gromacs trajectory file to add to the visualisation state file.")
                        )
    parser.add_argument("-vs", default=True, action="store_false", dest='virtual_sites',
                        help=("Don't write bonds between virtual sites and their constructing atoms. "
                              " (Bonds are written by default. Specify this flag if you don't want them written.)")
                        )
    parser.add_argument("-el", default=False,




                        action="store_true", dest='elastic',
                        help="Write elastic network of input proteins to separate files"
                        )
    parser.add_argument("-go", default=False, action="store_true",
                        help="Go network options")
    parser.add_argument("-gf", type=Path, dest='go_path',
                        help="Nonbonded parameter itp file for your go network")
    parser.add_argument("-vf", default=False, action="store_true",
                        help="Write out associated Visualisation Files (cg_bonds, vis.vmd) in the present directory")
    parser.add_argument("-pf", default=False, action="store_true",
                        help=("Write out the Protein File (protein.vmd) "
                             "for varying the bead size in proteins in addition to vis.vmd"))
    parser.add_argument("-ext", default=False, action="store_true",
                        help="Write system bonds to text files instead of topology files. Useful for non-VMD visualisation.")
    parser.add_argument("-cyl", default=False, action="store_true",
                        help="Write a tcl script to write out all network bonds for static images")

    args = parser.parse_args()

    if (args.cyl) and (args.system is None):
        raise ValueError("If you want a cylinder drawing tcl file"
                         " you must also give the coordinate file")

    if args.cyl:
        args.elastic = True

    ff, topol_lines, system_defines = system_reading(args.topology)

    written_mols, mol_bonds, mol_lens = molecule_editor(ff, topol_lines,
                                                        virtual_sites=args.virtual_sites,
                                                        ext=args.ext,
                                                        elastic=args.elastic,
                                                        go=args.go)

    if args.elastic:
        topol_writing(topol_lines, written_mols, 'en', w_include=args.system)
    if args.go:
        topol_writing(topol_lines, written_mols, 'go', w_include=args.system)
    topol_writing(topol_lines, written_mols, w_include=args.system)

    if args.system is not None:
        index_writing(args.system)
        if args.cyl:
            cylinders(args.system, topol_lines, mol_bonds, mol_lens)

    if args.vf:
        for file in os.listdir(DATA_PATH):
            if os.path.isfile(os.path.join(DATA_PATH, file)):
                if file == 'proteins.vmd' and not args.pf:
                    continue
                shutil.copy(os.path.join(DATA_PATH, file),
                            os.path.join(os.getcwd(), file))

        if args.elastic:
            if args.cyl:
                output_file_append(args.system, args.trajectory, args.pf, cylinders=True)
            else:
                output_file_append(args.system, args.trajectory, args.pf, elastic=True)
        if args.go:
            output_file_append(args.system, args.trajectory, args.pf, go=True)

    print('All done!')

if __name__ == '__main__':
    main()



# Copyright 2024 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vermouth.gmx import read_itp
from vermouth.forcefield import ForceField
from .topology import input_topol_reader
import re


def output_str(pairs):
    op_str = '{'
    for i in pairs:
        op_str += f'\u007b{i[0]} {i[1]}\u007d '
    return op_str[:-1] + '}'


def secondary_structure_parsing(lines, molname):

    header = []
    # this should ensure we only get the header
    for line in lines:
        if line[0] == ';':
            # ignore the ; at the start and the \n at the end
            header.append(line[1:-1].strip())
        else:
            break

    ss_string = ''
    # this should get the code
    for line in header:
        if line.upper() == line:
            ss_string = line

    helices = [i.span() for i in re.finditer('([H|G|I]{3,})', ss_string)]
    sheets = [i.span() for i in re.finditer('([B|E]{3,})', ss_string)]

    vmd_excl_str = ''
    for i in helices+sheets:
        vmd_excl_str += f'(resid > {i[0]} and resid < {i[1]}) or '
    vmd_excl_str = vmd_excl_str[:-4]

    hlx_col_str = ''
    for i in helices:
        hlx_col_str += f'(resid > {i[0]} and resid < {i[1]}) or '
    hlx_col_str = hlx_col_str[:-4]

    sht_col_str = ''
    for i in sheets:
        sht_col_str += f'(resid > {i[0]} and resid < {i[1]}) or '
    sht_col_str = sht_col_str[:-4]

    if len(helices) > 2 or len(sheets) > 2:
        with open(f'{molname}_cgsecstruct.txt', 'w') as f:
            f.write("suggested commands for viewing you molecule with cg_bonds-v6.tcl:\n")
            f.write(f'cg_helix {output_str(helices)} -hlxcolor "purple" -hlxfilled yes -hlxrad 3 -hlxmethod cylinder -hlxmat "AOChalky" -hlxres 50\n')
            f.write(f'cg_sheet {output_str(sheets)} -shtfilled "yes" -shtmat "AOChalky" -shtres 50 -shtcolor "red" -shtmethod flatarrow -shtarrwidth 5 -shtheadsize 10 -shtarrthick 3 -shtsides "sharp"\n')
            f.write('\nAdditionally, use the following command to remove the BB string from your molecule:\n')
            f.write(f'name BB and not ({vmd_excl_str})')
            f.write('\nThese commands will have to be modified to specify molid if you have multiple proteins in your system\n')
            f.write('\nAlternatively use the following to just colour the backbone:')
            f.write(f'\nhelices: name BB and ({hlx_col_str})')
            f.write(f'\nsheets: name BB and ({sht_col_str})')


def _misc_file_reader(lines):
    """
    Try to handle miscellaneous files which can't be read into the force field by read_itp

    Parameters
    ----------
    lines: list
        list of lines from the input .itp file


    Returns
    -------
    defines: dict
        dictionary of definition: parameters for bonded terms that have been externally defined
    others: list
        lines from an input itp that don't have #define in them

    if others == lines, None is returned, because the input file doesn't contain anything interesting
    """
    defines = {i.split()[1]: i.split(';')[0].split()[2:5] for i in lines if '#define' in i}
    others = [line for line in lines if '#define' not in line]
    if others == lines:
        return None
    else:
        return defines, others


def system_reading(topology):

    """
    read a .top file's contents into a ForceField
    """

    # get the topology file
    print(f"Reading input {topology}")
    topol_lines = input_topol_reader(topology)

    # for each molecule in the system, read in the itp
    d = {}
    for i, j in enumerate(topol_lines['core_itps']):
        with open(j, encoding="utf-8") as f:
            d[i] = f.readlines()

    # read the molecules into the forcefield
    ff = ForceField('martini3001')

    system_defines = {}
    for i, j in enumerate(d.keys()):
        try:
            read_itp(d[j], ff)
            secondary_structure_parsing(d[j], list(ff.blocks)[-1])
        except OSError:
            '''
            if we can't read the file into the system directly, we have something that isn't strictly a molecule
            most likely its the force field definition file (eg. martini_v3.0.0.itp) but we can't be sure
            a common one is something with #defines in for generic molecule bonded terms
            '''
            misc_result = _misc_file_reader(d[j])
            if misc_result is None:
                # if [ defaults ] is found, then it's martini_v3.0.0.itp or similar. ignore it.
                if "[ defaults ]" not in [k.strip() for k in d[j]]:
                    print(f"Error reading {topol_lines['core_itps'][i]}. Will ignore and exclude from output system.")
            else:
                # this means we've separated things out successfully and can read the actual itp content now
                # tracking system_defines is useful for when we sort the #TODOs below.
                for key, value in misc_result[0].items():
                    system_defines[key] = value
                read_itp(misc_result[1], ff)

    return ff, topol_lines, system_defines
